/* $Workfile: A66FxoFxs.c $       $Revision: 10 $      $Modtime: 9/04/09 4:47p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    A66FxoFxs.c

    A basic example console application demonstrating the use of OTX Driver
    with the Arni-6x6-PCI or the Arni-6x6-PCI board.

    The sample application configures 6 FXO POTS Line Interfaces and FXS line.
    The application provides a simple console interface to manage the POTS 
    Line Interface (LI) devices.
   
    The user can use the following simple keystrokes as commands to go 
    off-hook and on-hook, to dial numbers, etc.

    In this basic demo the FXO POTS LIs are supposed to be connected to a  
    telephone line or a telephone line simulator.
    
    The main command keys are listed in the following:
     - Press 'h' or '?' for help. 'q' to quit.
     

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxBrd.h"
#include "OtxCst.h"
#include "OtxDsp.h"
#include "OtxBurst.h"
#include "OtxBurstD.h"
#include "OtxSlic.h"



/*******************************************************************************
  Macros, Types
*******************************************************************************/

#define DEMO_DSP_MAX 1
#define DEMO_FXS_MAX 6
#define DEMO_LI_MAX  1

#define DEMO_TASK_REF           1L                   // Task Reference for a non-blocking call
#define DEMO_DEFAULT_LOG_FILE   "A66FxoFxs.log"      // Name of the LOG file
#define DEMO_DEFAULT_BOARD_NO   0
#define DEMO_DEFAULT_FXO_NO     2
#define DEMO_NO_LI_CONNECTED    0xFF

#define DEMO_BURST_BUF_SIZE     0x8000    // 32k

#define DEMO_U_LAW_SILENCE   0xFF    // Sending this byte produces silence in u-law mode
#define DEMO_A_LAW_SILENCE   0xD5    // Sending this byte produces silence in a-law mode

//#define DEMO_USE_DTMF_DIALER  // in lieu of Burst sender

#define DEMO_HWY_DSP0   0
#define DEMO_HWY_DSP1   1
#define DEMO_HWY_DSP2   2
#define DEMO_HWY_DSP3   3
#define DEMO_HWY_DSP4   4
#define DEMO_HWY_DSP5   5
#define DEMO_HWY_BURST  8


typedef struct DemoEnvT {
  char           *pszInputStr;
  OTX_BOOL        bExitOnError;
  char           *logFileName;
  char           *outFileName;
  int             boardNo;
  int             fxoNo;
  OtxPhysicalDeviceTypeE  nBoardType;
  OtxLogicalDeviceTypeE  eDetectorType;
  OtxDataFormatE  law;
  OtxCstPotsCfgSetE  nSetId;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
//static void DemoDisconnectFxoToFxsDevices(void);
static void DemoConnectFxoToFxsDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static char DemoDtmfParamToChar(int nParam);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static OTX_INT32 DemoProcessUserCommand(char commandChar);
static int DemoExit(int nExitCode);
static OTX_INT32 DemoSlicReceiveConfigMenu(char commandChar);

/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue        = OTX_INVALID_HANDLE_VALUE;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
FILE       *g_fStreamFile = NULL;
OTX_BOOL    g_bWatchDogThreadActive = OTX_FALSE;
OTX_UINT32  g_liConNumber_1 = DEMO_NO_LI_CONNECTED;
OTX_UINT32  g_liConTimeSlot = 0;
OTX_UINT8   g_rawDigit;

static OsaMutexS g_demoLogfCritSec;

struct PhysDevs {  // Handles to Physical Devices
  OTX_HANDLE hBoardController0;
  OTX_HANDLE hDsp[6];
  OTX_HANDLE hSlicLi[DEMO_FXS_MAX];
  OTX_HANDLE hBurst;
  OTX_HANDLE hTss;
} g_hPhysDevs;


// direct Burst DMA send
struct sBurstSender {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  OTX_UINT8   sqTxBuf[DEMO_BURST_BUF_SIZE];
  FILE        *pTxFile;
  OTX_INT32   nTsToSend;  
  OTX_UINT32  cbTxBuf;
  OTX_BOOL    bFileLoop;
};

// direct Burst DMA receive
struct sBurstReceiver {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  FILE        *pRxFile;
  OTX_INT32   nTsToRecord;  
};

// Pots Li device
struct sPotsLi {
  OTX_BOOL    bOffHook;
};


struct LogicDevs {  // Handles to Logical Devices
  OTX_HANDLE hPotsLi[DEMO_LI_MAX];

  OTX_HANDLE hVaDetector[DEMO_LI_MAX];
  OTX_HANDLE hCptDetector[DEMO_LI_MAX];
/*  
  OTX_HANDLE hDtmfDetector[DEMO_LI_MAX];
  OTX_HANDLE hDtmfDialer[DEMO_LI_MAX];
  OTX_HANDLE hVoiceMixer[DEMO_LI_MAX];
*/
  OTX_HANDLE hUserApp[DEMO_LI_MAX];     // only used for debugging of the DSP program
} g_hLogicDevs;

struct BurstDevs {
  struct sBurstSender mBurstSender;        
  struct sBurstReceiver mBurstReceiver;    
} g_sBurstDevs;

struct PotsDevs {
  struct sPotsLi mPotsLi;        
} g_sPotsDevs[DEMO_LI_MAX];


struct Pipes {  // Handles to Pipes
  OTX_HANDLE hPotsRx[DEMO_LI_MAX];
  OTX_HANDLE hPotsTx[DEMO_LI_MAX];
} g_hPipes;

char *g_sqNumber0 = "102";

// Call states for Slic device
enum DemoSlicDeviceCallState { 
  MAKE_BUSY                   = OTX_SLIC_CS_END_OF_STATES + 0 , 
  STATE_CHANGE                = OTX_SLIC_CS_END_OF_STATES + 1,
  INITIALIZING                = OTX_SLIC_CS_END_OF_STATES + 2,
  MAKE_RING_BACK              = OTX_SLIC_CS_END_OF_STATES + 3,
  MAKE_REORDER                = OTX_SLIC_CS_END_OF_STATES + 4, 
  MAKE_CONGESTION             = OTX_SLIC_CS_END_OF_STATES + 5,
  CALLBACK_PROGRAMMED         = OTX_SLIC_CS_END_OF_STATES + 6,
  CALL_BACK                   = OTX_SLIC_CS_END_OF_STATES + 7,
  MAKE_OFFHOOK                = OTX_SLIC_CS_END_OF_STATES + 8,
  ONHOOK                      = OTX_SLIC_CS_END_OF_STATES + 9,
  OFFHOOK                     = OTX_SLIC_CS_END_OF_STATES + 10,
  FLASHHOOK                   = OTX_SLIC_CS_END_OF_STATES + 11,
  DIGIT_DECODING              = OTX_SLIC_CS_END_OF_STATES + 12,
  LOOP_TRANSITION             = OTX_SLIC_CS_END_OF_STATES + 13, 
  CALLER_ID                   = OTX_SLIC_CS_END_OF_STATES + 14,
  RINGING                     = OTX_SLIC_CS_END_OF_STATES + 15,
  STOP_RINGING                = OTX_SLIC_CS_END_OF_STATES + 16,
  DTMF_TRANISTION             = OTX_SLIC_CS_END_OF_STATES + 17,
  NO_ACTION_NEEDED            = OTX_SLIC_CS_END_OF_STATES + 18
};
typedef enum DemoSlicDeviceCallState DemoSlicDeviceCallState;


// Call states for Slic device
char *g_sqszSlicDeviceCallState[] = { 
  // OTX Driver states
  "(drv)INITIALIZING                "    ,
  "(drv)CALIBRATE                   "    ,
  "(drv)MAKE_OFFHOOK                "    ,
  "(drv)ONHOOK                      "    ,
  "(drv)OFFHOOK                     "    ,
  "(drv)FIRST_RING                  "    ,
  "(drv)DEFERED_CALLBACK_PROGRAMMED "    ,
  "(drv)CALLER_ID                   "    ,
  "(drv)RINGING                     "    ,
  "(drv)DTMF_TRANSITION             "    ,
                                    
   // Demo App states               
  "(app)MAKE_BUSY                   "    ,
  "(app)STATE_CHANGE                "    ,
  "(app)INITIALIZING                "    ,
  "(app)MAKE_RING_BACK              "    ,
  "(app)MAKE_REORDER                "    ,
  "(app)MAKE_CONGESTION             "    ,
  "(app)CALLBACK_PROGRAMMED         "    ,
  "(app)CALL_BACK                   "    ,
  "(app)MAKE_OFFHOOK                "    ,
  "(app)ONHOOK                      "    ,
  "(app)OFFHOOK                     "    ,
  "(app)FLASHHOOK                   "    ,
  "(app)DIGIT_DECODING              "    ,
  "(app)LOOP_TRANSITION             "    ,
  "(app)CALLER_ID                   "    ,
  "(app)RINGING                     "    ,
  "(app)STOP_RINGING                "    ,
  "(app)DTMF_TRANISTION             "    ,
  "(app)NO_ACTION_NEEDED            "    ,
                                   
};

/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************
  @func
    Create the event queue to be used by the driver to inform the application
    about events, such as line status changes, or detected DMTF digits.
*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

}



void DemoOpenPhysicalDevices(void)
/*******************************************************************************
  @func
    Opens the Physical Devices on the Arni-16-PCI board

  @comm
    Calls function <f OtxDrvOpenPhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Get Handle to the Physical ARNI Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_ARNI_6X6_PCI_PLUS:
    DemoLogf("Opening Arni-6x6-PCI-Plus board\n");
    break;
  case OTX_DEVICE_ARNI_6X3_PCI_PLUS:
    DemoLogf("Opening Arni-6x3-PCI-Plus board\n");
    break;
  case OTX_DEVICE_ARNI_6X6_PCI:
    DemoLogf("Opening Arni-6x6-PCI board\n");
    break;
  case OTX_DEVICE_ARNI_6_FXO_PCI_PLUS:
  case OTX_DEVICE_ARNI_6_FXS_PCI_PLUS:
    DemoLogf("Opening Arni-6-FXO-PCI-Plus or Arni-6-FXS-PCI-Plus board\n");
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", 
             g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hBoardController0); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Check if the watchdog timer has triggered before we continue
  {
    OTX_UINT32 readVal;
    result = OtxDrvReadReg32(g_hPhysDevs.hBoardController0, 0x04, &readVal); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    if (readVal == 0) {
      DemoLogf("The watchdog timer has triggered. A reboot of the PC is necessary.\n");
      DemoExitQuestion();
    }
  }

  // Disable the Watchdog Timer, and enable it when we have completed the configuration
  result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_ENABLED,  OTX_FALSE); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Get a handle to the Time-Space Switch (TSS) on the board (just for debugging purposes)
  DemoLogf("Open OTX_DEVICE_TSS\n");
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_TSS, 0, 0, g_hEventQueue, &g_hPhysDevs.hTss); DemoCheckRc(g_hPhysDevs.hBoardController0, result);


  for (i = 0; i < DEMO_DSP_MAX; i++) {
    DemoLogf("Open OTX_DEVICE_DSP #%d\n", g_demoEnv.fxoNo);
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_DSP, g_demoEnv.fxoNo, g_demoEnv.fxoNo, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

  for (i = 0; i < DEMO_DSP_MAX; i++) {
    if (g_demoEnv.outFileName) {
      DemoLogf("Loading program file %s to DSP#%d \n", g_demoEnv.outFileName, g_demoEnv.fxoNo);
      result = OtxDspRunProgram(g_hPhysDevs.hDsp[i], g_demoEnv.outFileName, "_c_int00"); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    } else {
      DemoLogf("Loading OTX_DSP_SPM_OTXCST2 to DSP#%d \n", g_demoEnv.fxoNo);
      result = OtxDspRunSpm(g_hPhysDevs.hDsp[i], OTX_DSP_SPM_OTXCST2); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    }
  }

  // Get a handle to the physical Burst device.
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_BURST, 0, 0, g_hEventQueue, &g_hPhysDevs.hBurst); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // set device buffer event trigger levels
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,  DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs.hBurst, result);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,    DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs.hBurst, result);

  // Open SLICs
  for (i = 0 ; i < DEMO_FXS_MAX; i++) {
    DemoLogf("Open OTX_DEVICE_LI_SLIC  #%d \n", i);
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, 
                                      OTX_DEVICE_LI_SLIC, 
                                      i, 
                                      i, 
                                      g_hEventQueue, 
                                      &g_hPhysDevs.hSlicLi[i]); 
    DemoCheckRc(g_hPhysDevs.hSlicLi[i], result);
    
    // Take the fail-over relays out of the pass-through state 
    // (i.e. engage the SLIC on the line)
    DemoLogf("Take the fail-over relay out of "
             "the pass-through state for SLIC #%d\n", i);
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hSlicLi[i], 
                                         OTX_ATTR_SLIC_RELAY_CONTROL, 
                                         OTX_TRUE); 
    DemoCheckRc(g_hPhysDevs.hSlicLi[i], result);    
  }

#if DEMO_DSP_MAX == 1
  {
    OTX_HANDLE hTemp;
    // Only support FXO 2 at this time. Move SLIC 2 handle to index 0;

    hTemp = g_hPhysDevs.hSlicLi[0];
    g_hPhysDevs.hSlicLi[0] = g_hPhysDevs.hSlicLi[2];
    g_hPhysDevs.hSlicLi[2] = hTemp;
  }
#endif

}



void DemoEnablePhysicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Physical Devices
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  DemoLogf("In DemoEnablePhysicalDevices\n");
  result = OtxDrvEnable(g_hPhysDevs.hBoardController0, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  for (i = 0; i < DEMO_DSP_MAX; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

#if 1
  for (i = 0; i < DEMO_FXS_MAX; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hSlicLi[i], OTX_TASK_SYNC, 5000L); 
    DemoCheckRc(g_hPhysDevs.hSlicLi[i], result);
  }
#else
    i = 0; // Only support FXO 2 at this time

    result = OtxDrvEnable(g_hPhysDevs.hSlicLi[i], OTX_TASK_SYNC, 5000L); 
    DemoCheckRc(g_hPhysDevs.hSlicLi[i], result);
  
#endif
}


/*******************************************************************************
  @func
    Create the needed Logical Devices

  @comm
    The logical devices are different from the phyiscal devices in that the
    the number of logical devices is not fixed. More of them can be requested
    by the user and allocated by the system if needed.
*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  nDsp;
  OTX_INT32  nLi;

  for (nLi = 0; nLi < DEMO_LI_MAX; nLi++) {
    
    nDsp = nLi;
    
    DemoLogf("Creating Logical Devices on Dsp%d\n", nDsp);
    
    if (g_demoEnv.outFileName) {
      // For debug prints
      result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nLi], OTX_LDEVICE_USER_APPLICATION, g_hEventQueue, nLi, &g_hLogicDevs.hUserApp[nLi]); 
      //DemoCheckRc(g_hPhysDevs.hDsp[nLi], result);
      DemoLogf("Created User App device [%d], Handle = 0x%X (for debug printfs)\n", nLi, g_hLogicDevs.hUserApp[nLi]);
    }
    
    // Create Pots Li Logical Devices
    //DemoLogf("Create Logical Device OTX_LDEVICE_CST_LI_POTS #%d (nDsp=%d)\n", nLi, nDsp);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_CST_LI_POTS, g_hEventQueue, nLi, &g_hLogicDevs.hPotsLi[nLi]); DemoCheckRc(g_hLogicDevs.hPotsLi[nLi], result);

    // Create Voice Activity Detector Logical Devices
    //DemoLogf("Create Logical Device OTX_LDEVICE_CST_VOICE_ACTIVITY_DETECTOR #%d (nDsp=%d)\n", nLi, nDsp);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_CST_VOICE_ACTIVITY_DETECTOR, g_hEventQueue, nLi, &g_hLogicDevs.hVaDetector[nLi]); DemoCheckRc(g_hLogicDevs.hVaDetector[nLi], result);

    // Create Cpt Detector Logical Devices
    //DemoLogf("Create Logical Device OTX_LDEVICE_CST_CPT_DETECTOR #%d (nDsp=%d)\n", nLi, nDsp);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_CST_CPT_DETECTOR, g_hEventQueue, nLi, &g_hLogicDevs.hCptDetector[nLi]); DemoCheckRc(g_hLogicDevs.hCptDetector[nLi], result);

    
  } 
  
}


void DemoConnectDevices(
/*******************************************************************************
  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.
*******************************************************************************/
  void
)
{
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_INT32 nTs;

  g_liConNumber_1 = 0;
  DemoConnectFxoToFxsDevices();
  g_liConNumber_1 = DEMO_NO_LI_CONNECTED;

#if 1 // For additional debugging
  {
    OTX_HANDLE hPipeRecDaa;
    OTX_INT32 nHwy;
    
    // Also connect the DAA samples to the burst via a TSS connection
    result = OtxDrvCreatePipe(64, &hPipeRecDaa);
    DemoCheckRc(hPipeRecDaa, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = 0;
    nHwy = 2;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeRecDaa, OTX_PIPE_INPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeRecDaa, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = 0;
    nHwy = 8;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeRecDaa, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeRecDaa, result);
  }
#endif


#if 1 // For additional debugging
  {
    OTX_HANDLE hPipeRecSlic;
    OTX_INT32 nHwy;
  
    // Also connect the FXS SLIC(2) samples to the burst via a TSS connection
    result = OtxDrvCreatePipe(64, &hPipeRecSlic);
    DemoCheckRc(hPipeRecSlic, result);
  
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = 6;
    nHwy = 6;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeRecSlic, OTX_PIPE_INPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeRecSlic, result);
  
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = 1;
    nHwy = 8;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeRecSlic, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeRecSlic, result);
  }
#endif


}


static void DemoConnectFxoToFxsDevices( void )
/*******************************************************************************
  @func
    Create the needed Pipe Devices and connect the physical and 
    logical devices to pipes.

    This function connects FXO to FXS. If FXS handset 2 is taken off hook
    the FXO 2 will be connected. Each FXS (SLIC) is directly mapped to the
    it corresponding FXO POTS line.

    All six of the SLIC devices use Physical Highway 6, 
    DSP highways are 0-5 with DSP zero connected to highway 0.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream 
    (a subset of a Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. 
    A pipe is a characterized by its capacity (in 8 kbit/s increments). 
    The user can create a pipe with the driver function <f OtxDrvCreatePipe>.
    A pipe is connected to physical or logical devices and mapped to
    specific time-slots with the driver function <f OtxDrvConnectPipe>.
    
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_INT32 nTs;

  // This Demo connect two devices together
  DemoLogf("DemoConnectFxoToFxsDevices\n");


  // Data from and to POTS LIs
  if(DEMO_NO_LI_CONNECTED == g_liConNumber_1)  {
    // The demo has a sequence issue no connects are set
    DemoLogf("Warning DemoConnectLiToLiDevices: No connections were set!\n");
  }
  else {
    result = OtxDrvCreatePipe(64, &g_hPipes.hPotsRx[g_liConNumber_1]); 
    DemoCheckRc(g_hPipes.hPotsRx[g_liConNumber_1], result);
    result = OtxDrvCreatePipe(64, &g_hPipes.hPotsTx[g_liConNumber_1]); 
    DemoCheckRc(g_hPipes.hPotsTx[g_liConNumber_1], result);

    DemoLogf("  Time slot %d\n", g_liConTimeSlot);

    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = g_liConTimeSlot;
    sqTimeSlotMask[nTs] = 0xFF;


    // Connect Pipes
    // Connect FXO X to the FXS X as an input
    result = OtxDrvConnectPipe(g_hPipes.hPotsRx[g_liConNumber_1], 
                               OTX_PIPE_OUTPUT,  
                               g_hPhysDevs.hSlicLi[g_liConNumber_1], 
                               6,                  // Physical Highway, 
                               sqTimeSlotMask,     // sqTimeSlotMask
                               32);                // Size of connection mask
    DemoCheckRc(g_hPipes.hPotsRx[g_liConNumber_1], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = 0;
    sqTimeSlotMask[nTs] = 0xFF;

    // Connect FXO X to the FXS X as an output
    result = OtxDrvConnectPipe(g_hPipes.hPotsRx[g_liConNumber_1], 
                               OTX_PIPE_INPUT,  
                               g_hPhysDevs.hDsp[g_liConNumber_1], 
                               2,                  // Physical Highway  (only for DSP 2)
                               sqTimeSlotMask,     // sqTimeSlotMask
                               32);                // Size of connection mask
    DemoCheckRc(g_hPipes.hPotsRx[g_liConNumber_1], result);


    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = g_liConTimeSlot;
    sqTimeSlotMask[nTs] = 0xFF;


    // Connect FXO X to the FXS X as an output
    result = OtxDrvConnectPipe(g_hPipes.hPotsTx[g_liConNumber_1],
                               OTX_PIPE_INPUT,  
                               g_hPhysDevs.hSlicLi[g_liConNumber_1], 
                               6,                  // Physical Highway
                               sqTimeSlotMask,     // sqTimeSlotMask
                               32);                // Size of connection mask
    DemoCheckRc(g_hPipes.hPotsTx[g_liConNumber_1], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask));
    nTs = 2;
    sqTimeSlotMask[nTs] = 0xFF;

    // Connect FXO X to the FXS X as an input
    result = OtxDrvConnectPipe(g_hPipes.hPotsTx[g_liConNumber_1],
                               OTX_PIPE_OUTPUT,  
                               g_hPhysDevs.hDsp[g_liConNumber_1], 
                               2,                  // Physical Highway (only for DSP 2)
                               sqTimeSlotMask,     // sqTimeSlotMask
                               32);                // Size of connection mask
    DemoCheckRc(g_hPipes.hPotsTx[g_liConNumber_1], result);




    // CONNECT LOGICAL DEVICES WITHIN THE DSP
    // --------------------------------------


    // Create the following Connections in receive direction:
    //   SLIC (FXS)  --> Pipe --> Pots Li (FXO)
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hPotsLi[g_liConNumber_1], 
                                        -1, 
                                        g_hPipes.hPotsRx[g_liConNumber_1], 
                                        -1); 
    DemoCheckRc(g_hPipes.hPotsTx[g_liConNumber_1], result);


    // Make the following connections in transmit direction:
    //   Pots Li (FXO)  --> Pipe --->  SLIC (FXS) 
    result = OtxDrvConnectLogicalDevice(g_hPipes.hPotsTx[g_liConNumber_1], 
                                        -1, 
                                        g_hLogicDevs.hPotsLi[g_liConNumber_1], 
                                        -1); 
    DemoCheckRc(g_hLogicDevs.hPotsLi[g_liConNumber_1], result);  
  }

  DemoLogf("Done\n");

  return;
}

#if 0
static void DemoDisconnectFxoToFxsDevices( void )
/*******************************************************************************
  @func
    Disconnect a pipe

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream 
    (a subset of a Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. 
    A pipe is a characterized by its capacity (in 8 kbit/s increments). 
    The user can create a pipe with the driver function <f OtxDrvCreatePipe>.
    A pipe is connected to physical or logical devices and mapped to
    specific time-slots with the driver function <f OtxDrvConnectPipe>.
    
*******************************************************************************/
{
  OTX_RESULT result;

  // This Demo connected two devices together
  DemoLogf("DemoDisconnectFxoToFxsDevices\n");

  if( (DEMO_NO_LI_CONNECTED == g_liConNumber_1) )  {
    DemoLogf("   No Li were connected\n");
  }
  else {
    // Data from and to POTS LIs
    result = OtxDrvDisconnectPipe(g_hPipes.hPotsRx[g_liConNumber_1], OTX_PIPE_OUTPUT); 
    DemoCheckRc(g_hPipes.hPotsRx[g_liConNumber_1], result);
    result = OtxDrvDisconnectPipe(g_hPipes.hPotsRx[g_liConNumber_1], OTX_PIPE_INPUT); 
    DemoCheckRc(g_hPipes.hPotsRx[g_liConNumber_1], result);
    result = OtxDrvDisconnectPipe(g_hPipes.hPotsTx[g_liConNumber_1], OTX_PIPE_OUTPUT); 
    DemoCheckRc(g_hPipes.hPotsTx[g_liConNumber_1], result);
    result = OtxDrvDisconnectPipe(g_hPipes.hPotsTx[g_liConNumber_1], OTX_PIPE_INPUT); 
    DemoCheckRc(g_hPipes.hPotsTx[g_liConNumber_1], result);

    // Set to unused flag
    g_liConNumber_1 = DEMO_NO_LI_CONNECTED;
  }

  DemoLogf("Done\n");

  return;
}
#endif


void DemoInitializeLogicalDevices(
/*******************************************************************************
  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.
*******************************************************************************/
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  DemoLogf("In DemoInitializeLogicalDevices\n");


  DemoLogf("Configuring Lis for country set=");
  switch (g_demoEnv.nSetId) {
  case OTX_CST_POTS_CFG_SET_USA:
    DemoLogf("USA");
    break;
  case OTX_CST_POTS_CFG_SET_GERMANY: 
    DemoLogf("Germany");
    break;
  case OTX_CST_POTS_CFG_SET_UK:
    DemoLogf("United Kingdom");
    break;
  case OTX_CST_POTS_CFG_SET_FRANCE:
    DemoLogf("France");
    break;
  case OTX_CST_POTS_CFG_SET_AUSTRALIA:
    DemoLogf("Australia");
    break;
  case OTX_CST_POTS_CFG_SET_JAPAN:
    DemoLogf("Japan");
    break;
  default:
    DemoLogf("UNKNOWN");
    break;
  }
  DemoLogf("\n");  

  for (i = 0; i < DEMO_LI_MAX; i++) {

    // Initialize the POTS LI logical devices
    result = OtxCstPotsInit(
      g_hLogicDevs.hPotsLi[i],
      g_demoEnv.law,
      g_demoEnv.nSetId
      );
    DemoCheckRc(g_hLogicDevs.hPotsLi[i], result);


    // Initialize the Voice Activity Detector
    result = OtxCstVoiceActivityDetectorInit(
      g_hLogicDevs.hVaDetector[i], 
      g_demoEnv.law
      ); 
    DemoCheckRc(g_hLogicDevs.hVaDetector[i], result);

    // Initialize the CPT Detector
    result = OtxCstCptDetectorInit(
      g_hLogicDevs.hCptDetector[i], 
      g_demoEnv.law, 
      OTX_TRUE,
      OTX_CST_CPT_DETECTOR_Q35_CFG
      ); 
    DemoCheckRc(g_hLogicDevs.hCptDetector[i], result);

  }
}




void DemoEnableLogicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Logical Devices
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  DemoLogf("In DemoEnableLogicalDevices\n");

  for (i = 0; i < DEMO_LI_MAX; i++) {

    result = OtxDrvEnable(g_hLogicDevs.hPotsLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hPotsLi[i], result);


    result = OtxDrvEnable(g_hLogicDevs.hCptDetector[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hCptDetector[i], result);

    result = OtxDrvEnable(g_hLogicDevs.hVaDetector[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hCptDetector[i], result);

  }

}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX [%d.%d])\n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result));
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    DemoExit(1);
  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
    
}


/*******************************************************************************
  @func
    Function that keeps reseting the Watchdog timer on the board
*******************************************************************************/
void  *DemoWatchDogFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_BOOL bEnabled;

  g_bWatchDogThreadActive = OTX_TRUE;

  DemoLogf("DemoWatchDogFunction Thread started\n");

  while (g_bWatchDogThreadActive) {

    result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_ENABLED,  &bEnabled); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    if (!bEnabled) {
      DemoLogf("*********** Watchdog Timer timed out. Application must be restarted *********\n");
      OsaThreadSleep(3000); // in milliseconds
      continue;
    }

    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_RESET,  OTX_TRUE /* Any value will work */); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    OsaThreadSleep(100); // in milliseconds
  }

  DemoLogf("DemoWatchDogFunction Thread terminating\n");
  return 0;

}




/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OTX_INT32 i;
  OTX_UINT32 hookStatus;
  OTX_UINT32 slicState;
  OtxEventDataS eventData;
  OTX_RESULT result;

  for(i = 0; i < DEMO_LI_MAX; i++) {
    OtxSlicGetHookState( g_hPhysDevs.hSlicLi[i], &hookStatus );
    DemoLogf("Slic[%d] Hook Status %d \n", i, hookStatus);
  }

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); 
      DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {

        switch (eventData.m_eDeviceType) {

          case OTX_DEVICE_LI_SLIC:
            // Determine the event
            switch (eventData.m_nCode) {
              // Pulse Metering Active Timer Interrupt
              case (OTX_SLIC_PULSE_INACTIVE):
              // Pulse Metering Inactive Timer Interrupt
              case (OTX_SLIC_PULSE_ACTIVE):
              // Ring Trip Interrupt
              case (OTX_SLIC_RING_TRIP):  
              // Loop Closure Transition Interrupt
              case (OTX_SLIC_LOOP_STATE):  
              // Indirect Register Access Serviced
              case (OTX_SLIC_INDIRECT_ACCESS_RDY):
              // Common Mode Calibration Error
              case (OTX_SLIC_ERROR_CM_CAL):  
              // Continue Ringing Active Timer
              case (OTX_SLIC_RING_INACTIVE_CONT):  
              // Oscillator 1 Active Timer Interrupt
              case (OTX_SLIC_OSC1_INACTIVE):
              // Oscillator 1 Inactive Timer Interrupt
              case (OTX_SLIC_OSC1_ACTIVE):  
              // Oscillator 2 Active Timer Interrupt
              case (OTX_SLIC_OSC2_INACTIVE):
              // Oscillator 2 Inactive Timer Interrupt
              case (OTX_SLIC_OSC2_ACTIVE): 
              // Ringing Active Timer Interrupt
              case (OTX_SLIC_RING_INACTIVE):
              // Ringing Inactive Timer Interrupt.
              case (OTX_SLIC_RING_ACTIVE):  
                // Nothing more to do here
                break;

              // Power Alarm Q1 Interrupt
              case (OTX_SLIC_PWR_ALARM_PQ1):  
              // Power Alarm Q2 Interrupt
              case (OTX_SLIC_PWR_ALARM_PQ2):  
              // Power Alarm Q3 Interrupt
              case (OTX_SLIC_PWR_ALARM_PQ3):  
              // Power Alarm Q4 Interrupt
              case (OTX_SLIC_PWR_ALARM_PQ4):  
              // Power Alarm Q5 Interrupt
              case (OTX_SLIC_PWR_ALARM_PQ5):  
              // Power Alarm Q6 Interrupt
              case (OTX_SLIC_PWR_ALARM_PQ6):
                DemoLogf("    Slic[%d] Power Alarm!\n", eventData.m_nSourceId);
                break;

              // DTMF Tone Detected
              case (OTX_SLIC_DTMF_DETECT):
                // DTMF digit is passed with event
                g_rawDigit = (OTX_UINT8) eventData.m_nParam;
                DemoLogf("    Slic[%d] DTMF digit detected %X\n", 
                         eventData.m_nSourceId,
                         g_rawDigit);

                break;

              // Slic indicates OFFHOOK condition
              case (OTX_SLIC_OFF_HOOK):  


                // SLIC is off hook setup connection FSO to FXS
                // For now only FXO 2 is supported
                if( (eventData.m_nSourceId == DEMO_DEFAULT_FXO_NO) &&   
                    (g_liConNumber_1       == DEMO_NO_LI_CONNECTED) ) {
                  DemoLogf("    Slic[%d] Off Hook phone making connection\n", 
                           eventData.m_nSourceId);

#if DEMO_DSP_MAX != 1
                  g_liConNumber_1 = eventData.m_nSourceId;
                  // Each slic will have a fixed location within the data stream.
                  // Slic0 will be at TS0 and TS1, 
                  // Slic1 will be at TS3 and TS4... Slic5 will be at TS15 and TS16
                  g_liConTimeSlot = (OTX_UINT16) ((eventData.m_nSourceId * 3));
#else
                  // only one LI supported at this time
                  g_liConNumber_1 = 0;
                  g_liConTimeSlot = (OTX_UINT16) ((DEMO_DEFAULT_FXO_NO * 3));
#endif

                  // Connect slic audio for device starting the call
                  result = OtxSlicEnableAudio( g_hPhysDevs.hSlicLi[g_liConNumber_1], 
                                               &slicState );  
                  DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);

                  if(slicState == OTX_SLIC_CS_ONHOOK) {
                    DemoLogf("Warning: Driver state Li%d OTX_SLIC_CS_ONHOOK\n", 
                             g_liConNumber_1);
                  } 
                  
//                  DemoConnectFxoToFxsDevices();

                  // Set Active Pots Li Off-Hook ...
                  DemoLogf("    Pots Li#%d Off-hook\n", eventData.m_nSourceId);
                  result = OtxCstPotsOffHook(g_hLogicDevs.hPotsLi[g_liConNumber_1], 
                                             0x1000, 
                                             2000); 
                  DemoCheckRc(g_hLogicDevs.hPotsLi[g_liConNumber_1], result);
                
                
                
                }
                else {
                  if(g_liConNumber_1 == DEMO_NO_LI_CONNECTED) {
                    // Work around for FXO
                    DemoLogf("    Slic[%d] Only FXO 2 is supported at this time!\n", 
                             eventData.m_nSourceId);
                  }
                  else {
                    DemoLogf("    Slic[%d] Off Hook: phone already connected\n", 
                             eventData.m_nSourceId);
                  }
                }
                
                break;
              // Slic indicates ONHOOK condition
              case (OTX_SLIC_ON_HOOK):

                if( eventData.m_nSourceId != DEMO_NO_LI_CONNECTED ) {
                  // There was a connection clean it up
//                  DemoDisconnectFxoToFxsDevices();

                  // Set Active Pots Li On-Hook ...
                  DemoLogf("    Pots Li#%d On-hook\n", eventData.m_nSourceId);
                  result = OtxCstPotsOnHook(g_hLogicDevs.hPotsLi[g_liConNumber_1], 
                                            0x1001, 
                                            2000); 
                  DemoCheckRc(g_hLogicDevs.hPotsLi[g_liConNumber_1], result);


                }
                g_liConNumber_1 = DEMO_NO_LI_CONNECTED;
                
                break;

            default:
              DemoLogf("Slic[%d] Unknown Event: %ld \n", 
                       eventData.m_nSourceId, 
                       eventData.m_nCode);

              break;
            }

            break;

          default:
        
            // We have a driver event
            // Printf the event data.
            // In a real application branch based on the data
            DemoPrintEvent(&eventData);
            break;
        }    
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;

}


/*******************************************************************************
  @func
    Only used by the USER_APPLICATION device for debugging of DSP program code.
*******************************************************************************/
void DemoPrintCSTEvent(OtxEventDataS *pEventData) 
{
  OTX_RESULT otxRc;
  char msgStr[256];
  
  if (pEventData->m_nCode == OTX_USER_APPLICATION_EC_READ_DATA) {
    OTX_UINT8  buf[8192];
    OTX_UINT32 nBytesReturned;
    OTX_UINT32 nIoControlCode;      
    
    otxRc = OtxDspReadData(pEventData->m_hDevice, 
                           OTX_TASK_SYNC, 
                           buf, 
                           8192, 
                           &nBytesReturned, 
                           &nIoControlCode);
    if (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
      switch(nIoControlCode) {
      case 0x2ff:
        {
          char strBuf[1024];
          int k;
          sprintf(strBuf, "Len=%d :",nBytesReturned);
          for(k=0;k<(int)nBytesReturned;k+=2){
            sprintf(&strBuf[0] + strlen(strBuf), "%c",buf[k]);
          }
          sprintf(&strBuf[0] + strlen(strBuf), "\n");
          DemoLogf(strBuf);
        }
        break;
        
      case 0x0200 /*DEMO1_NOTIFY_DEBUG_DATA*/:
        {
          DemoLogf("DEMO1_NOTIFY_DEBUG_DATA (nRequestId=%d, nBytes=%d)\n", 
                   pEventData->m_nRequestId, nBytesReturned);
        }
        break;

      case 0x02FE:
        {
          static int counter = 0;

          if (counter++ % 5000 == 0) {
            DemoLogf("Got data\n");
          }

          if (g_fStreamFile) {
            fwrite((OTX_UINT8 *)buf, nBytesReturned, sizeof(OTX_UINT8), g_fStreamFile);
          }

        }
        break;
      default:
        DemoLogf("Unknown IoControlCode received in OtsDspReadData"
                 " (nIoControlCode = %d, nBytesReturned=%d, nRequestId=%d)\n", 
                 nIoControlCode, 
                 nBytesReturned, 
                 pEventData->m_nRequestId);
        break;
      }
    } else {
      OTX_RESULT errorCode = otxRc;
      otxRc = OtxDrvResultCode2String(pEventData->m_hDevice, errorCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("User App[%d] Error: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("User App[%d] Invalid Error Code: %ld\n", pEventData->m_nSourceId, errorCode);
      }
      
    }
    
  } else {
    
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, 
                                   pEventData->m_nCode, 
                                   256, 
                                   &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("User App[%d] Event: %s (nRequestId=%d)\n", 
               pEventData->m_nSourceId, 
               msgStr, 
               pEventData->m_nRequestId);
    } else {
      DemoLogf("User App[%d] Invalid Event: %ld (nRequestId=%d)\n", 
               pEventData->m_nSourceId, 
               pEventData->m_nCode, 
               pEventData->m_nRequestId);
    }
  }
}




/*******************************************************************************
  @func
    Burst Sender Function
*******************************************************************************/
void DemoBurstSender(OTX_UINT32 nSourceId)
{
  OTX_RESULT  otxRc;
  OTX_UINT32  ui32;


  // validate parameters
  if (nSourceId != 0) {
    DemoLogf("DemoBurstSender: Invalid nSourceId(%d)!\n", nSourceId);
    return;
  }

  if (!g_sBurstDevs.mBurstSender.bActive) {
    // we're not active, so just return
    return;
  }

  if (g_sBurstDevs.mBurstSender.pTxFile) {
    // Prime the transmit buffer with 0xFF (only really needed for single TS transmission)
    memset(&g_sBurstDevs.mBurstSender.sqTxBuf[0], 0xFF, sizeof(g_sBurstDevs.mBurstSender.sqTxBuf) / sizeof(g_sBurstDevs.mBurstSender.sqTxBuf[0]));

    if (g_sBurstDevs.mBurstSender.nTsToSend == -1) {
      // All TS
      g_sBurstDevs.mBurstSender.cbTxBuf = fread(
        &g_sBurstDevs.mBurstSender.sqTxBuf,
        sizeof(g_sBurstDevs.mBurstSender.sqTxBuf[0]),
        sizeof(g_sBurstDevs.mBurstSender.sqTxBuf) / sizeof(g_sBurstDevs.mBurstSender.sqTxBuf[0]),
        g_sBurstDevs.mBurstSender.pTxFile
      );

    } else {
      // Single TS
      OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE / 32];
      int cbSingleTsBuf;
      int byteCount = 0;
      int x;
      
      cbSingleTsBuf = fread(
        &singleTsBuf,
        sizeof(singleTsBuf[0]),
        sizeof(singleTsBuf) / sizeof(singleTsBuf[0]),
        g_sBurstDevs.mBurstSender.pTxFile
        );

      
      // Dispurse the single TS to be transmitted in the whole transmit buffer 
      byteCount = 0;
      for ( x = g_sBurstDevs.mBurstSender.nTsToSend; 
      (byteCount < cbSingleTsBuf) && (x < DEMO_BURST_BUF_SIZE); 
      x += 32
        ) {
        g_sBurstDevs.mBurstSender.sqTxBuf[x] = singleTsBuf[byteCount++];
      }
      
      assert(byteCount == cbSingleTsBuf);
      g_sBurstDevs.mBurstSender.cbTxBuf = cbSingleTsBuf * 32;

      if (!feof(g_sBurstDevs.mBurstSender.pTxFile)) {
        assert(g_sBurstDevs.mBurstSender.cbTxBuf == DEMO_BURST_BUF_SIZE);
      } else {
       DemoLogf("DemoBurstSender[%d]: Reached EOF\n", nSourceId);
      }

    }
        
    if (g_sBurstDevs.mBurstSender.cbTxBuf > 0) {

      DemoLogf("DemoBurstSender[%d]: Transmitting %d bytes\n", nSourceId, g_sBurstDevs.mBurstSender.cbTxBuf);

      ui32 = g_sBurstDevs.mBurstSender.cbTxBuf;
      otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, g_sBurstDevs.mBurstSender.sqTxBuf, &ui32);
      DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

      // check the amount buffered w/ how much we requested to be buffered
      //  (since we set the trigger amount to the buffer size, there should
      //    have been room to buffer this, so these two should be the same).
      assert(ui32 == g_sBurstDevs.mBurstSender.cbTxBuf);

    } else {

      if (g_sBurstDevs.mBurstSender.bFileLoop) {
        if (g_sBurstDevs.mBurstSender.bVerbose) {
          DemoLogf("DemoBurstSender[%d]: Transmission complete. Starting file from the beginning.\n", nSourceId);
        }
        rewind(g_sBurstDevs.mBurstSender.pTxFile);
      } else {
        DemoLogf("DemoBurstSender[%d]: Transmission complete. Closing file\n", nSourceId);
        fclose(g_sBurstDevs.mBurstSender.pTxFile);
        g_sBurstDevs.mBurstSender.pTxFile = NULL;
        g_sBurstDevs.mBurstSender.bActive = OTX_FALSE;
      }
    }

  } else {
    // Send a text string (continously until aborted)

    // we MUST round amount to send to be bounded on the number of TS the highway hosts.
    //  (the buffer is representative of a burst DMA buffer for all TS on the highway hosted, TS0-TSn)

    g_sBurstDevs.mBurstSender.cbTxBuf = g_sBurstDevs.mBurstSender.cbTxBuf;
    assert(g_sBurstDevs.mBurstSender.cbTxBuf == DEMO_BURST_BUF_SIZE);

    ui32 = g_sBurstDevs.mBurstSender.cbTxBuf;
    otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, g_sBurstDevs.mBurstSender.sqTxBuf, &ui32);
    DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

    // check the amount buffered w/ how much we requested to be buffered
    //  (since we set the trigger amount to the buffer size, there should
    //    have been room to buffer this, so these two should be the same).
    assert(ui32 == g_sBurstDevs.mBurstSender.cbTxBuf);
  }

  return;
}

/*******************************************************************************
  @func
    Burst Receiver Function
*******************************************************************************/
void DemoBurstReceiver(OTX_UINT32 nSourceId)
{
  OTX_RESULT    otxRc;
  OTX_UINT8     buf[DEMO_BURST_BUF_SIZE];
  OTX_UINT32    ui32;
  OTX_UINT64    nFrameRefId;
  OTX_DATETIME  dtFrameRefTime;

  // validate parameters
  if (nSourceId < 0 || nSourceId > 2) {
    DemoLogf("DemoBurstReceiver: Invalid nSourceId(%d)...\n", nSourceId);
    return;
  }

  if (!g_sBurstDevs.mBurstReceiver.bActive) {
    // we're not active, so just return
    return;
  }

  // we want the frame reference id and time
  ui32 = sizeof(buf);
  otxRc = OtxBurstReadDataEx(g_hPhysDevs.hBurst, nSourceId, buf, &ui32, &nFrameRefId, &dtFrameRefTime);
  DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

  // check the amount buffered w/ how much we requested to be buffered
  //  (since we set the trigger amount to the buffer size, there should
  //    have been room to buffer this, so these two should be the same).
  assert(ui32 == sizeof(buf));

  // echo if we're in verbose mode
  if (g_sBurstDevs.mBurstReceiver.bVerbose) {
    int x;

    DemoPrintTimeStamp(dtFrameRefTime);

    // note: we only display the least significant 32 bits of the frame reference id.
    DemoLogf("BurstReceiver[%d] (len=%.5d) ID:%08x ", nSourceId, sizeof(buf), (OTX_UINT32) (nFrameRefId & 0xffffffff));

    for (x = 0; (x < 32) && (x < (OTX_INT32) sizeof(buf)); x++) {
      // display the highway's currentLi's data
      if ((x % 8) == 0) DemoLogf("\n");
      DemoLogf("%.2X ", buf[x]);
    }
    DemoLogf("\n");
    DemoLogf("\n");
  }

  // write data to a file if desired
  if (g_sBurstDevs.mBurstReceiver.pRxFile) {
    int cbWritten;

    if (g_sBurstDevs.mBurstReceiver.nTsToRecord == -1) {
      // All TS
      cbWritten = fwrite(buf, sizeof(buf[0]), sizeof(buf), g_sBurstDevs.mBurstReceiver.pRxFile);
      if (cbWritten != (int) sizeof(buf)) {
        DemoLogf("BurstReceiver[%d] Error: Could not write data to file.  (nFrameLen=%d, cbWritten=%d) \n", nSourceId, sizeof(buf), cbWritten);
      }
    } else {
      // Single TS
      OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE/32];
      int byteCount = 0;
      int x;

      byteCount = 0;
      for (x = g_sBurstDevs.mBurstReceiver.nTsToRecord; x < (OTX_INT32) sizeof(buf); x += 32) {
        singleTsBuf[byteCount++] = buf[x];
      }
      cbWritten = fwrite(singleTsBuf, sizeof(singleTsBuf[0]), byteCount, g_sBurstDevs.mBurstReceiver.pRxFile);
      if (cbWritten != byteCount) {
        DemoLogf("BurstReceiver[%d] Error: Could not write data to file.  (byteCount=%d, cbWritten=%d) \n", nSourceId, byteCount, cbWritten);
      }
    }
  }
}


/*******************************************************************************

  @func

  Convert a CPT (Call Progress Tone) Detection parameter to a string.

  @comm

*******************************************************************************/
static char *DemoCptParamToStr(int nParam) 
{
  static char strBuf[128];

  switch (nParam) {

  case OTX_CST_CPT_DETECTOR_NO_TONE:        
    return "Silence, No tone";
  case OTX_CST_CPT_DETECTOR_DIAL_TONE:      
    return "Dial Tone";
  case OTX_CST_CPT_DETECTOR_RING_TONE:      
    return "Ring Tone";
  case OTX_CST_CPT_DETECTOR_BUSY_TONE:      
    return "Busy Tone";
  case OTX_CST_CPT_DETECTOR_FASTBUSY_TONE:  
    return "FastBusy Tone";
  case OTX_CST_CPT_DETECTOR_LONG_TONE:      
    return "Long lasting tone for 1 second";
  case OTX_CST_CPT_DETECTOR_END_OF_TONE:    
    return "End of current tone";
  default: // Error
    // DemoLogf("Unrecognized CPT tone parameter (%d)\n", nParam);
    sprintf(strBuf, "Unrecognized CPT Parameter (nParam=0x%X)", nParam);
    return strBuf;
  }
}



void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  //DemoPrintTimeStamp(pEventData->m_nTimeStamp);

  switch (pEventData->m_eDeviceType) {
    
  case OTX_LDEVICE_USER_APPLICATION:
    // Events from the DspDemo1.out logical devices
    DemoPrintCSTEvent(pEventData);
    break;

  case OTX_DEVICE_DSP:
    {
      // Non-logical device related event from the DSPs
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("DSP[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
    
  case OTX_LDEVICE_CST_DTMF_DIALER:
    {
      switch (pEventData->m_nCode) {
      case OTX_CST_DTMF_DIALER_EC_COMPLETE:
        DemoLogf("DtmfDialer[%d] Completed DTMF dialing of %d digits\n", pEventData->m_nSourceId, pEventData->m_nParam);
        break;
      default:
        DemoLogf("DtmfDialer[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
        break;
      }
    }
    break;


  case OTX_LDEVICE_CST_CPT_DETECTOR:
    {
      // Event from Cpt Detector
      switch (pEventData->m_nCode) {
      case OTX_CST_CPT_DETECTOR_EC_MTONESTART:
        {
          DemoLogf("CptDetector[%d] MTONESTART, %s\n", pEventData->m_nSourceId, DemoCptParamToStr(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_CPT_DETECTOR_EC_MTONE:
        {
          DemoLogf("CptDetector[%d] MTONE, Tone confirmed, %s\n", pEventData->m_nSourceId, DemoCptParamToStr(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_CPT_DETECTOR_EC_MSIGNALEND:
        {
          DemoLogf("CptDetector[%d] MSIGNALEND, %s\n", pEventData->m_nSourceId, DemoCptParamToStr(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_CPT_DETECTOR_EC_MFAIL:
        {
          DemoLogf("CptDetector[%d] MFAIL, %s\n", pEventData->m_nSourceId, DemoCptParamToStr(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_CPT_DETECTOR_EC_MPATTERNEND:
        {
          DemoLogf("CptDetector[%d] MPATTERNEND, %s\n", pEventData->m_nSourceId, DemoCptParamToStr(pEventData->m_nParam));
        }
        break;
        
      default:
        DemoLogf("CptDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
        break;
        
      }
    }
    break;

  case OTX_LDEVICE_CST_DTMF_DETECTOR:
    {
      // Event from Dtmf Detector
      switch (pEventData->m_nCode) {
      case OTX_CST_DTMF_DETECTOR_EC_MTONESTART:
        {
          DemoLogf("DtmfDetector[%d] MTONESTART, DTMF digit %c\n", pEventData->m_nSourceId, DemoDtmfParamToChar(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_DTMF_DETECTOR_EC_MTONE:
        {
          DemoLogf("DtmfDetector[%d] MTONE, DTMF tone confirmed\n", pEventData->m_nSourceId);
        }
        break;
        
      case OTX_CST_DTMF_DETECTOR_EC_MSIGNALEND:
        {
          DemoLogf("DtmfDetector[%d] MSIGNALEND, DTMF digit %c\n", pEventData->m_nSourceId, DemoDtmfParamToChar(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_DTMF_DETECTOR_EC_MFAIL:
        {
          DemoLogf("DtmfDetector[%d] MFAIL\n", pEventData->m_nSourceId);
        }
        break;
        
      case OTX_CST_DTMF_DETECTOR_EC_MPATTERNEND:
        {
          DemoLogf("DtmfDetector[%d] MPATTERNEND\n", pEventData->m_nSourceId);
        }
        break;
        
      default:
        DemoLogf("DtmfDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
        break;
        
      }
    }
    break;

  case OTX_LDEVICE_CST_VOICE_ACTIVITY_DETECTOR:
    {
      switch (pEventData->m_nCode) {
      case OTX_CST_VOICE_ACTIVITY_DETECTOR_EC_NOISE:
        DemoLogf("VoiceActivityDetector[%d] Noise present\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_VOICE_ACTIVITY_DETECTOR_EC_NOISE_HANGOVER:
        DemoLogf("VoiceActivityDetector[%d] Noise Hangover. Possibly turning into voice.\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_VOICE_ACTIVITY_DETECTOR_EC_VOICE:
        DemoLogf("VoiceActivityDetector[%d] Voice present\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_VOICE_ACTIVITY_DETECTOR_EC_VOICE_HANGOVER:
        DemoLogf("VoiceActivityDetector[%d] Voice Hangover. Had voice, but possibly turning into noice.\n", pEventData->m_nSourceId);
        break;
      default:
        DemoLogf("VoiceActivityDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
        break;
      }
    }
    break;


  case OTX_LDEVICE_CST_LI_POTS:
    {
      // Event from Pots Li
      switch (pEventData->m_nCode) {

      case OTX_CST_POTS_RING_SIGNAL_DETECTED:
        DemoLogf("PotsLi[%d] A ring signal has been detected\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_END_RING_SIGNAL_DETECTED:
        DemoLogf("PotsLi[%d] End of ring signal has been detected\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_AUTO_END_RING_SIGNAL_DETECTED:
        DemoLogf("PotsLi[%d] End of ring signal due to offhook has been detected\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_TIP_RING_REVERSAL_DETECTED:
        DemoLogf("PotsLi[%d] Tip/Ring reversal detected. The other end hung up upon us.\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_CALLER_ID_DETECTED:
        {
          char szCallerId[40];

          if (pEventData->m_nParam == OTX_CST_CALLER_ID_EC_NOERROR) {
            otxRc = OtxDrvGetAttributeValueSz(pEventData->m_hDevice, OTX_ATTR_CST_POTS_CALLER_ID, szCallerId, 40); 
            if (OTX_SUCCESS(otxRc)) {
              DemoLogf("PotsLi[%d] Caller Id = \"%s\"\n", pEventData->m_nSourceId, szCallerId);
            } else if (OTX_RESULT_CODE(otxRc) == OTX_RES_BUSY) {
              DemoLogf("PotsLi[%d] Caller Id Overrun. Caller Id string lost. Reseting the condition\n", pEventData->m_nSourceId);
              OtxDrvSetAttributeValueSz(pEventData->m_hDevice, OTX_ATTR_CST_POTS_CALLER_ID, NULL); 
            } else {
              DemoLogf("PotsLi[%d] Unexpected return code from OtxDrvGetAttributeValueSz (otxRc=0x%X)\n", pEventData->m_nSourceId, otxRc);
            }
          } else {
            DemoLogf("PotsLi[%d] Caller Id detection failure. nCode=%d\n", pEventData->m_nSourceId, pEventData->m_nParam);
          }

        }
        break;
      case OTX_CST_POTS_OFF_HOOK_COMPLETE:
        assert(pEventData->m_nSourceId < DEMO_LI_MAX);
        g_sPotsDevs[pEventData->m_nSourceId].mPotsLi.bOffHook = OTX_TRUE;
        DemoLogf("PotsLi[%d] Off-Hook command complete\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_ON_HOOK_COMPLETE:
        assert(pEventData->m_nSourceId < DEMO_LI_MAX);
        g_sPotsDevs[pEventData->m_nSourceId].mPotsLi.bOffHook = OTX_FALSE;
        DemoLogf("PotsLi[%d] On-Hook command complete\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_ON_HOOK_LINE_MON_COMPLETE:
        assert(pEventData->m_nSourceId < DEMO_LI_MAX);
        g_sPotsDevs[pEventData->m_nSourceId].mPotsLi.bOffHook = OTX_FALSE;
        DemoLogf("PotsLi[%d] On-Hook Line Monitor mode command complete\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_SENSE_LINE_COMPLETE:
        assert(pEventData->m_nSourceId < DEMO_LI_MAX);
        if (g_sPotsDevs[pEventData->m_nSourceId].mPotsLi.bOffHook) {
          // We are in the OffHook state so line sensing will return loop current (3mA per bit)
          DemoLogf("PotsLi[%d] Sense line command complete. Loop current = %d mA (%d) above minimum\n", pEventData->m_nSourceId, 3 * pEventData->m_nParam, pEventData->m_nParam);
        } else {
          double nVal = 2.75 * pEventData->m_nParam;
          // We are in the OnHook state so line sensing will return line voltage (2.75V per bit)
          DemoLogf("PotsLi[%d] Sense line command complete. Line voltage = %2.1f V  (%d)\n", pEventData->m_nSourceId, (double) nVal, pEventData->m_nParam);
        }
        break;
      case OTX_CST_POTS_PULSE_GEN_COMPLETE:
        DemoLogf("PotsLi[%d] Pulse Generator command complete\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_READ_REG_COMPLETE:
        DemoLogf("PotsLi[%d] ReadReg command complete Val=%d\n", pEventData->m_nSourceId, pEventData->m_nParam);
        break;
      case OTX_CST_POTS_WRITE_REG_COMPLETE:
        DemoLogf("PotsLi[%d] WriteReg command complete\n", pEventData->m_nSourceId);
        break;
      default:
        // Unexpected event from this device
        DemoLogf("PotsLi[%d] Unexpected Event (%d)\n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
    
  case OTX_DEVICE_BURST:
    {
      switch (pEventData->m_nCode) {
      case OTX_BURST_RECEIVER_EC_DATA_AVAILABLE:
        DemoBurstReceiver(pEventData->m_nRequestId);
        break;
      case OTX_BURST_SENDER_EC_BUF_DEPLETED:
        DemoBurstSender(pEventData->m_nRequestId);
        break;
      default:
        otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
        if (OTX_SUCCESS(otxRc)) {
          DemoLogf("Burst[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
        } else {
          DemoLogf("Burst[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
        }
        break;
      }
      
    }
    break;

  default:
    {
      // Unhandled event
      DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    }
    break;
    
  }
  
}

/*******************************************************************************

  @func

  Convert a DTMF parameter to a digit (char).

  @comm

*******************************************************************************/
static char DemoDtmfParamToChar(int nParam) {
  switch (nParam) {
  case 0:  // 0  (941Hz/1336Hz)
  case 1:  // 1  (697Hz/1209Hz)
  case 2:  // 2  (697Hz/1336Hz)
  case 3:  // 3  (697Hz/1477Hz)
  case 4:  // 4  (770Hz/1209Hz)
  case 5:  // 5  (770Hz/1336Hz)
  case 6:  // 6  (770Hz/1477Hz)
  case 7:  // 7  (852Hz/1209Hz)
  case 8:  // 8  (852Hz/1336Hz)
  case 9:  // 9  (852Hz/1477Hz)
    return '0' + nParam;
  case 10: // A  (697Hz/1633Hz)
  case 11: // B  (770Hz/1633Hz)
  case 12: // C  (852Hz/1633Hz)
  case 13: // D  (941Hz/1633Hz)
    return 'A' - 10 + nParam;
  case 14: // *  (941Hz/1209Hz)
    return '*';
    break;
  case 15: // #  (941Hz/1477Hz)
    return '#';
    break;
  default: // Error
    DemoLogf("Unrecognized DTMF tone parameter (%d)\n", nParam);
    return '-';
  }
}



/*******************************************************************************

  @func

    Returns the state of the Burst device (enabled or disabled).

*******************************************************************************/
OTX_BOOL DemoIsBurstDeviceEnabled()
{
  OTX_RESULT result;
  OtxDeviceStateE  eDevState;

  result = OtxDrvGetState(g_hPhysDevs.hBurst, &eDevState); 
  DemoCheckRc(g_hPhysDevs.hBurst, result);

  if (eDevState == OTX_DEVICE_WORKING) {
    return OTX_TRUE; // The burst device is enabled
  } else {
    return OTX_FALSE; // The burst device is not enabled
  }
}


OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************
  @func
    Processes User Commands
*******************************************************************************/
{
  OTX_RESULT        result;
  static OTX_INT32  nCurrentLi = 0;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd   Synopsis\n");
    DemoLogf("===   ===========================================================\n");
    DemoLogf(" ?    Display Help text\n");
    DemoLogf(" h    Ditto\n");
    DemoLogf(" q    Quit\n");

    //DemoLogf(" 0-5  Select active Line Interface\n");
    DemoLogf(" f    Set active Pots Li Off-Hook\n");
    DemoLogf(" n    Set active Pots Li On-Hook\n");
    DemoLogf(" m    Set active Pots Li in On-Hook Line-Monitor mode\n");
    DemoLogf(" c    Line sense. Read line voltage or loop current\n");

    DemoLogf(" l    Flash/hook on the active Pots Li\n");
    DemoLogf(" r    Read a Register from the DAA\n");
    DemoLogf(" w    Write to a Register in the DAA\n");
    DemoLogf(" b    Get Debug data from the DSP\n");
    DemoLogf(" s    Stream receive data to file (toggle). Only for DSP debugging\n");
    DemoLogf(" e    Enable/Disable Burst (toggle)\n");
    DemoLogf(" a    Abort Burst Senders and Receivers\n");
    DemoLogf(" R    Start Burst Receiver. Prompt for file to recv\n");
    DemoLogf(" !    Terminate Watchdog Timer Thread (Crash test)\n");
    DemoLogf(" v    Toggle Burst Receiver Verbose mode\n");
    DemoLogf(" z    SLIC Audio path configuration MENU\n");

    break;

  case 'v':
    {
      if (g_sBurstDevs.mBurstReceiver.bVerbose) {
        g_sBurstDevs.mBurstReceiver.bVerbose = OTX_FALSE;
      } else {
        g_sBurstDevs.mBurstReceiver.bVerbose = OTX_TRUE;
      }
      DemoLogf("Burst Receiver Verbose is now %s\n", g_sBurstDevs.mBurstReceiver.bVerbose ? "ON" : "OFF");
    }
    break;


  case '!':
    {
      DemoLogf("Terminating WatchDog Thread (Crash test)\n");
      g_bWatchDogThreadActive = OTX_FALSE;
    }
    break;


  case 'e':
    if (DemoIsBurstDeviceEnabled()) {
      // by disabling the Burst device, we stop the process of transferring data
      // from the hardware to the driver via DMA.
      DemoLogf("Disable g_hPhysDevs.hBurst\n");
      result = OtxDrvDisable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
      DemoCheckRc(g_hPhysDevs.hBurst, result);
    } else {
      // by enabling the Burst device, we start the process of transferring data
      // from the hardware to the driver via DMA.  The data is stored in the Burst
      // device's buffer for later processing to connected logical devices.
      DemoLogf("Enable g_hPhysDevs.hBurst\n");
      result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
      DemoCheckRc(g_hPhysDevs.hBurst, result);
    }
    break;

  case 'a':
    {
      if (DemoIsBurstDeviceEnabled()) {
        if (g_sBurstDevs.mBurstReceiver.pRxFile) {
          DemoLogf("DemoBurstReceiver: Closing file\n");
          fclose(g_sBurstDevs.mBurstReceiver.pRxFile);
          g_sBurstDevs.mBurstReceiver.pRxFile = NULL;
        }
        g_sBurstDevs.mBurstReceiver.bActive = OTX_FALSE;
        
        if (g_sBurstDevs.mBurstReceiver.pRxFile) {
          DemoLogf("mBurstSender: Closing file\n");
          fclose(g_sBurstDevs.mBurstSender.pTxFile);
          g_sBurstDevs.mBurstSender.pTxFile = NULL;
        }
        g_sBurstDevs.mBurstSender.bActive   = OTX_FALSE;
      } else {
        DemoLogf("The Burst device is not enabled. Command ignored\n");
      }
    }
    break;

  case 'R':
    {
      char sqzBuf[OTX_MAX_PATH];
      OTX_UINT32 nHighwayNo;
      static OTX_INT32 nTsToRecord = -1;

      if (DemoIsBurstDeviceEnabled()) {
        nHighwayNo = 0;
        if (g_sBurstDevs.mBurstReceiver.bActive == OTX_FALSE || 
            g_sBurstDevs.mBurstReceiver.pRxFile == NULL) {
          DemoLogf("BurstReceiver[%d]. Please enter file to receive: ", 
                   nHighwayNo);
          OsaConsoleGetStr(sqzBuf);
          if (sqzBuf[0] == '\0') {
            DemoLogf("  No file enterred abort capture.\n");
            break;
          }
          g_sBurstDevs.mBurstReceiver.pRxFile = fopen(sqzBuf, "wb");
          if (g_sBurstDevs.mBurstReceiver.pRxFile == NULL) {
            DemoLogf("Error: Could not open file %s \n", sqzBuf);
            break;
          }
          
          DemoLogf("BurstReceiver[%d].\n"
                   " Please enter single TS[0-31] to receive to file (-1=All): ", 
                   nHighwayNo);
          OsaConsoleGetStr(sqzBuf);
          if (sqzBuf[0] == '\0') {
            DemoLogf("  No TS enterred abort capture.\n");
            break;
          } else {
            nTsToRecord = strtol(sqzBuf, NULL, 0);
          }
          g_sBurstDevs.mBurstReceiver.nTsToRecord = nTsToRecord;
          
          g_sBurstDevs.mBurstReceiver.bActive = OTX_TRUE;
        } else {
          DemoLogf("Burst Receiver[%d] Thread already running.\n", 
                   nHighwayNo);
        }
      } else {
        DemoLogf("The Burst device is not enabled. Command ignored\n");
      }
    }
    break;

  case 'b':
    {
      DemoLogf("Debug Dump for Dsp#%d\n", nCurrentLi);
      OtxDrvDbgDump(g_hPhysDevs.hDsp[nCurrentLi], 0);
    }
    break;


  case 'f':
    // Set Active Pots Li Off-Hook ...
    DemoLogf("Pots Li#%d Off-hook\n", nCurrentLi);
    result = OtxCstPotsOffHook(g_hLogicDevs.hPotsLi[nCurrentLi], 0x1000, 2000); 
    DemoCheckRc(g_hLogicDevs.hPotsLi[nCurrentLi], result);
    break;


  case 'n':
    // Set Active Pots Li On-Hook ...
    DemoLogf("Pots Li#%d On-hook\n", nCurrentLi);
    result = OtxCstPotsOnHook(g_hLogicDevs.hPotsLi[nCurrentLi], 0x1001, 2000); 
    DemoCheckRc(g_hLogicDevs.hPotsLi[nCurrentLi], result);
    break;

  case 'm':
    // Set Active Pots Li in On-Hook Line Monitor mode ...
    DemoLogf("Pots Li#%d On-hook Line-Monitor mode\n", nCurrentLi);
    result = OtxCstPotsOnHookLineMon(g_hLogicDevs.hPotsLi[nCurrentLi], 0x1001, 2000); 
    DemoCheckRc(g_hLogicDevs.hPotsLi[nCurrentLi], result);
    break;

  case 'c':
    // Line Sense
    DemoLogf("Pots Li#%d Line Sense\n", nCurrentLi);
    result = OtxCstPotsSenseLine(g_hLogicDevs.hPotsLi[nCurrentLi], 0x1005, 2000); 
    DemoCheckRc(g_hLogicDevs.hPotsLi[nCurrentLi], result);
    break;

  case 'l':
    // Flash Hook
    DemoLogf("Pots Li#%d Hook Flash\n", nCurrentLi);
    result = OtxCstPotsHookFlash(g_hLogicDevs.hPotsLi[nCurrentLi], 200, 0x1002, 2000); DemoCheckRc(g_hLogicDevs.hPotsLi[nCurrentLi], result);
    break;

  case 'p':
    {
      // Pulse dial
      OTX_INT32 c;
      DemoLogf("Pots Li#%d Pulse Dialing. Press digit to dial... [0-9] ", nCurrentLi);
      c = OsaConsoleGetCh();
      DemoLogf("\n");
      result = OtxCstPotsPulseDial(g_hLogicDevs.hPotsLi[nCurrentLi], c, 0x1003, 2000); DemoCheckRc(g_hLogicDevs.hPotsLi[nCurrentLi], result);
      DemoLogf("Done.\n");
    }
    break;

#if 0 // Only one DSP/FXO is supported by this program at this time
  /*
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
    // Select current Li
    {
      OTX_INT32 nLi = commandChar - '0';
      
      nCurrentLi = nLi;
      DemoLogf("Li#%d Selected ....\n", nCurrentLi);
    }
    break;
  */
#endif

#if 1 // Only for DSP program debugging
  case 's':
    {
      char sqzBuf[OTX_MAX_PATH];
      
      if (g_fStreamFile) {
        fclose(g_fStreamFile);
        g_fStreamFile = NULL;
        DemoLogf("Closed streaming file\n");
      } else {
        
        DemoLogf("Please enter file to receive: ");
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0] == '\0') {
          break;
        }
        
        if ((g_fStreamFile = fopen(sqzBuf, "wb")) == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzBuf);
          break;
        }
      }
    }
    break;
#endif

#if 1 // Only for DSP program debugging
  case 'r':
    {
      OTX_UINT32 nCmd = 1; // 1 = Read
      static OTX_UINT8 nRegister = 2; 
      static OTX_UINT8 nDummy2 = 0; 
      static OTX_UINT8 nDummy3 = 0; 
      char strBuf[64];
      
      DemoLogf("Read DAA register. Register [%d]\n ", nRegister);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nRegister = (OTX_UINT8)strtol(strBuf, NULL, 0);
      }

      nCmd = (nDummy3 << 3*8) | (nDummy2 << 2*8) | (nRegister << 1*8) | nCmd;
      result = OtxDrvDbgDump(g_hLogicDevs.hPotsLi[nCurrentLi], nCmd);
      
    }
    break;
#endif

#if 1 // Only for DSP program debugging
  case 'w':
    {
      OTX_UINT32 nCmd = 2; // 2 = Write
      static OTX_UINT8 nRegister = 2; 
      static OTX_UINT8 nVal = 0; 
      static OTX_UINT8 nDummy3 = 0; 
      char strBuf[64];
      
      DemoLogf("Write DAA register. Register [%d]\n ", nRegister);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nRegister = (OTX_UINT8)strtol(strBuf, NULL, 0);
      }

      DemoLogf("Please enter value [0x%.2X] ", nVal);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nVal = (OTX_UINT8)strtol(strBuf, NULL, 0);
      }

      nCmd = (nDummy3 << 3*8) | (nVal << 2*8) | (nRegister << 1*8) | nCmd;
      result = OtxDrvDbgDump(g_hLogicDevs.hPotsLi[nCurrentLi], nCmd);
      
    }
    break;
#endif


  case 'z':
    DemoLogf("\nSLIC Receive Path Audio Configuration\n");

    {
      OTX_UINT32 exitMenu = 0;
      char strBuf[256];

      while( exitMenu == 0 ) {
        // Display help menu for user first
        exitMenu = DemoSlicReceiveConfigMenu('?');
        
        if(exitMenu == 0) {
          // Get user input
          OsaConsoleGetStr(strBuf);

          exitMenu = DemoSlicReceiveConfigMenu(strBuf[0]);
        }
      }

    }

    DemoLogf("\nAudio Configuration done\n");

    break;
    


  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;
      
      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);
      
      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
      
    }
    break;
    
  case 'q':
  case 'Q':
    // Quit ...
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}



/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->bExitOnError      = OTX_FALSE;
  envP->logFileName       = DEMO_DEFAULT_LOG_FILE;
  envP->outFileName       = NULL;
  envP->law               = OTX_DATA_FORMAT_COMPAND8_U_LAW;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->fxoNo             = DEMO_DEFAULT_FXO_NO;
  envP->nSetId            = OTX_CST_POTS_CFG_SET_USA;
  envP->nBoardType        = OTX_DEVICE_ARNI_6X6_PCI;
  envP->eDetectorType     = OTX_LDEVICE_CST_DTMF_DETECTOR;

  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          envP->logFileName = argv[i]+2;

        } else if (*(argv[i]+1) == 'O') {   // DSP out file
          envP->outFileName = argv[i]+2;

        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);

        } else if (*(argv[i]+1) == 'F') {   // FXO number
          envP->fxoNo = strtoul(argv[i]+2, NULL, 0);

        } else if (OsaToUpper(*(argv[i]+1)) == 'T') {   // Board Type
          if (OsaToUpper(*(argv[i] + 2)) == 'A') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X6_PCI_PLUS;
          } else if (OsaToUpper(*(argv[i] + 2)) == 'J') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X6_PCI;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'C') {   // Companding Law
          if (*(argv[i]+2) == 'A') {   // A-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_A_LAW;
          } else if (*(argv[i]+2) == 'U') {   // U-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_U_LAW;
          } else {
            DemoLogf("Error: Invalid companding law command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'S') {   // Configuration Set
          if (*(argv[i]+2) == 'U') {          // USA
            envP->nSetId = OTX_CST_POTS_CFG_SET_USA;
          } else if (*(argv[i]+2) == 'G') {   // Germany
            envP->nSetId = OTX_CST_POTS_CFG_SET_GERMANY;
          } else if (*(argv[i]+2) == 'E') {   // UK
            envP->nSetId = OTX_CST_POTS_CFG_SET_UK;
          } else if (*(argv[i]+2) == 'F') {   // France
            envP->nSetId = OTX_CST_POTS_CFG_SET_FRANCE;
          } else if (*(argv[i]+2) == 'A') {   // Australia
            envP->nSetId = OTX_CST_POTS_CFG_SET_AUSTRALIA;
          } else if (*(argv[i]+2) == 'J') {   // Japan
            envP->nSetId = OTX_CST_POTS_CFG_SET_JAPAN;
          } else {
            DemoLogf("Error: Unsupported country specific configuration set %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }

}


static void DemoHelpPanel()
/*******************************************************************************
  @func
    Displays a help for the command line.
*******************************************************************************/
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName>    : Log file name\n");
  DemoLogf(" /B<boardNo>     : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /F<fxoNo>       : FXO number to configure (Default is %d)\n", DEMO_DEFAULT_FXO_NO);
  DemoLogf(" /T[A|D|J]       : Board Type used for the demo, where\n");
  DemoLogf("                    A  = Use Arni-6-PCI-Plus Board\n");
  DemoLogf("                    D  = Use Arni-6x3-PCI-Plus Board\n");
  DemoLogf("                    J  = Use Arni-6-PCI Board\n");
  DemoLogf(" /C[A|U]         : Voice Compression Codec, where\n");
  DemoLogf("                    A  = A-law\n");
  DemoLogf("                    U  = u-law (default)\n");
  DemoLogf(" /S[U|G|E|F|A|J] : Country specific setting, where\n");
  DemoLogf("                    U  = Li Configuration for USA (default)\n");
  DemoLogf("                    G  = Li Configuration for Germany\n");
  DemoLogf("                    E  = Li Configuration for United Kingdom\n");
  DemoLogf("                    F  = Li Configuration for France\n");
  DemoLogf("                    A  = Li Configuration for Australia\n");
  DemoLogf("                    J  = Li Configuration for Japan\n");
  DemoLogf(" /O<fileName>    : DSP Out file name (for debugging only)\n");
}


static void DemoLogf(char *formatStr, ...)
/*******************************************************************************
  @func
    Displays a message on the screen and prints it to the logfile.
*******************************************************************************/
{
  va_list argList;

  OsaMutexLock(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  OsaMutexUnlock(&g_demoLogfCritSec);
}



/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  //OTX_INT32  i;
  OTX_INT32  exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char       commandChar;

	OsaConsoleInit();

  OsaMutexInit(&g_demoLogfCritSec);

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    exit(-1);
  }

  DemoLogf("%s, %s, Odin TeleSystems Inc.\n", argv[0], __DATE__);
  DemoLogf("Arni-6x6-PCI, Basic FXO Demo Program.\n");

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);
  DemoLogf("FXO   number %d\n", g_demoEnv.fxoNo);
  if (g_demoEnv.outFileName) {
    DemoLogf("Out file to be loaded is %s\n", g_demoEnv.outFileName);
  }

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    DemoExit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    DemoExit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }


  // Create Event Queues for Driver --> Application Communications
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Arni-16-PCI board
  DemoOpenPhysicalDevices();

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);

    free(pszCpldBuildNumbers);

  }

  // Enable Physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnablePhysicalDevices();

  OsaThreadStart(DemoIdleFunction);

  // Create and get handles to the Logical devices needed by this application
  DemoCreateLogicalDevices();

  // Create and get handles to the Pipe devices needed by this application
  // Connect Pipes to Physical and Logical Devices
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoInitializeLogicalDevices();

  // Enable Logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnableLogicalDevices();

#if 0 // Disable watchdog while debugging
  // Enable the Watchdog Timer itself
  result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_ENABLED,  OTX_TRUE); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Start the Watchdog timer Reset Thread, which will keep the watchdog alive
  OsaThreadStart(DemoWatchDogFunction);
#endif  

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....

  DemoLogf("\nPress 'h' for help, 'q' to quit ...\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
		exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Delete allocated resources
  DemoExit(0);

  return 0;
}

/*******************************************************************************
*******************************************************************************/
static int DemoExit(int nExitCode)
{

  DemoLogf("\nExiting...\n");

  // Important to disable the Watchdog Timer since we will not be reseting it after we exit
  // the application. 
  OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, 
                              OTX_ATTR_BRD_WATCHDOG_TIMER_ENABLED,  
                              OTX_FALSE);

  g_bWatchDogThreadActive = OTX_FALSE;

  if (DemoIsBurstDeviceEnabled()) {
    if (g_sBurstDevs.mBurstReceiver.pRxFile) {
      DemoLogf("DemoExit: Closing burst file\n");
      fclose(g_sBurstDevs.mBurstReceiver.pRxFile);
      g_sBurstDevs.mBurstReceiver.pRxFile = NULL;
    }
    g_sBurstDevs.mBurstReceiver.bActive = OTX_FALSE;
    
    if (g_sBurstDevs.mBurstReceiver.pRxFile) {
      DemoLogf("DemoExit: Closing burst file\n");
      fclose(g_sBurstDevs.mBurstSender.pTxFile);
      g_sBurstDevs.mBurstSender.pTxFile = NULL;
    }
    g_sBurstDevs.mBurstSender.bActive   = OTX_FALSE;
  }

  OsaThreadSleep(100);   // give threads a chance to exit

  // Clean up
  OsaConsoleExit();
  OtxDrvReleaseLib();

  exit(nExitCode);
}



/*******************************************************************************

  @func

  Processes User Commands for SLIC audio path setup

  @comm

  This code is case sensitive.

*******************************************************************************/
OTX_INT32 DemoSlicReceiveConfigMenu(char commandChar)
{
  char strBuf[256];
//  OTX_UINT32 regReadAddL = 0;
  OTX_UINT32 regValueL = 0;
  OTX_RESULT result;
//  unsigned char regValue = 0;

  if(g_liConNumber_1 == DEMO_NO_LI_CONNECTED) {
      DemoLogf("\nAudio Path Configuration Requires FXO to FXS Connection\n\n");
      return 1;
  }

  switch(commandChar)   {

  case 'a':

    DemoLogf("Audio Receive Path Gain\n");

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                            9,     // Audio gain control register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);

    DemoLogf("\nRegister value was %#X\n", regValueL);
    // Bit 0 and 1 are used to set Analog receive path gain
    regValueL = regValueL & 0x03;

    switch(regValueL)   {

      case 0:
        DemoLogf("Gain set to 0 dB\n");
      break;
      case 1:
        DemoLogf("Gain set to -3.5 dB\n");
      break;
      case 2:
        DemoLogf("Gain set to 3.5 dB\n");
      break;
      case 3:
        DemoLogf("Analog receive path muted\n");
      break;

      default:
        DemoLogf("Software application error. Unsupported option.\n");
      return 0;
    }

    DemoLogf("Enter Option \n");
    DemoLogf("0: for O db Gain\n");
    DemoLogf("1: for -3.5 db Gain\n");
    DemoLogf("2: for 3.5 db Gain\n");
    DemoLogf("3: Analog receive path muted\n");
    DemoLogf("> ");
    OsaConsoleGetStr(strBuf);

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                            9,     // Audio gain control register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);
    regValueL = regValueL & ~(0x03);

    switch(strBuf[0])   {

      case '0':
        DemoLogf("Gain set to 0 dB\n");
      break;
      case '1':
        DemoLogf("Gain set to -3.5 dB\n");
        regValueL = regValueL | 0x01;
      break;
      case '2':
        DemoLogf("Gain set to 3.5 dB\n");
        regValueL = regValueL | 0x02;
      break;
      case '3':
        DemoLogf("Analog receive path muted\n");
        regValueL = regValueL | 0x03;
      break;

      default:
        DemoLogf("Software application error. Unsupported option.\n");
      return 0;
    }

    result = OtxDrvWriteReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                             9,     // Audio gain control register
                             regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);


    
    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                            9,     // Audio gain control register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);

    DemoLogf("Register value is %#X\n", regValueL);

    DemoLogf("Done\n");

  break;


  case 'A':

    DemoLogf("Audio Transmit Path Gain\n");

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                            9,     // Audio gain control register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);

    DemoLogf("\nRegister value was %#X\n", regValueL);
    // Bit 0 and 1 are used to set Analog receive path gain
    regValueL = regValueL & 0x0C;

    switch(regValueL)   {

      case 0:
        DemoLogf("Gain set to 0 dB\n");
      break;
      case 0x04:
        DemoLogf("Gain set to -3.5 dB\n");
      break;
      case 0x08:
        DemoLogf("Gain set to 3.5 dB\n");
      break;
      case 0x0C:
        DemoLogf("Analog transmit path muted\n");
      break;

      default:
        DemoLogf("Software application error. Unsupported option.\n");
      return 0;
    }

    DemoLogf("Enter Option \n");
    DemoLogf("0: for O db Gain\n");
    DemoLogf("1: for -3.5 db Gain\n");
    DemoLogf("2: for 3.5 db Gain\n");
    DemoLogf("3: Analog receive path muted\n");
    DemoLogf("> ");
    OsaConsoleGetStr(strBuf);

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                            9,     // Audio gain control register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);
    regValueL = regValueL & ~(0x0C);

    switch(strBuf[0])   {

      case '0':
        DemoLogf("Gain set to 0 dB\n");
      break;
      case '1':
        DemoLogf("Gain set to -3.5 dB\n");
        regValueL = regValueL | 0x04;
      break;
      case '2':
        DemoLogf("Gain set to 3.5 dB\n");
        regValueL = regValueL | 0x08;
      break;
      case '3':
        DemoLogf("Analog transmit path muted\n");
        regValueL = regValueL | 0x0C;
      break;

      default:
        DemoLogf("Software application error. Unsupported option.\n");
      return 0;
    }

    result = OtxDrvWriteReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                             9,     // Audio gain control register
                             regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);


    
    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                            9,     // Audio gain control register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);

    DemoLogf("Register value is %#X\n", regValueL);

    DemoLogf("Done\n");

  break;

  case 'b':

    DemoLogf("Audio Hybrid Gain\n");

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                            11,     // Audio Hybrid gain control register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);

    DemoLogf("\nRegister value was %#X\n", regValueL);
    // Bit 0, 1, and 2 are used to set Hybrid gain
    regValueL = regValueL & 0x07;

    switch(regValueL)   {

      case 0:
        DemoLogf("Gain set to 4.08 dB\n");
      break;
      case 1:
        DemoLogf("Gain set to 2.5 dB\n");
      break;
      case 2:
        DemoLogf("Gain set to 1.16 dB\n");
      break;
      case 3:
        DemoLogf("Gain set to 0 dB\n");
      break;
      case 4:
        DemoLogf("Gain set to -1.02 dB\n");
      break;
      case 5:
        DemoLogf("Gain set to -1.94 dB\n");
      break;
      case 6:
        DemoLogf("Gain set to -2.77 dB\n");
      break;
      case 7:
        DemoLogf("Audio Hybrid off\n");
      break;

      default:
        DemoLogf("Software application error. Unsupported option.\n");
      return 0;
    }

    DemoLogf("Enter Option \n");
    DemoLogf("0: for 4.08 db Gain\n");
    DemoLogf("1: for 2.5 db Gain\n");
    DemoLogf("2: for 1.16 db Gain\n");
    DemoLogf("3: for O db Gain\n");
    DemoLogf("4: for -1.02 db Gain\n");
    DemoLogf("5: for -1.94 db Gain\n");
    DemoLogf("6: for -2.77 db Gain\n");
    DemoLogf("7: Audio Hybrid off\n");
    DemoLogf("> ");
    OsaConsoleGetStr(strBuf);

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                            11,     // Audio Hybrid gain control register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);

    regValueL = regValueL & ~(0x07);

    switch(strBuf[0])   {

      case '0':
        DemoLogf("Gain set to 4.08 dB\n");
      break;
      case '1':
        DemoLogf("Gain set to 2.5 dB\n");
        regValueL = regValueL | 0x01;
      break;
      case '2':
        DemoLogf("Gain set to 1.16 dB\n");
        regValueL = regValueL | 0x02;
      break;
      case '3':
        DemoLogf("Gain set to 0 dB\n");
        regValueL = regValueL | 0x03;
      break;
      case '4':
        DemoLogf("Gain set to -1.02 dB\n");
        regValueL = regValueL | 0x04;
      break;
      case '5':
        DemoLogf("Gain set to -1.94 dB\n");
        regValueL = regValueL | 0x05;
      break;
      case '6':
        DemoLogf("Gain set to -2.77 dB\n");
        regValueL = regValueL | 0x06;
      break;
      case '7':
        DemoLogf("Audio Hybrid off\n");
        regValueL = regValueL | 0x07;
      break;

      default:
        DemoLogf("Software application error. Unsupported option.\n");
      return 0;
    }

    result = OtxDrvWriteReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                             11,     // Audio Hybrid gain control register
                             regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);
    
    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                            11,     // Audio Hybrid gain control register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);

    DemoLogf("Register value is %#X\n", regValueL);

    DemoLogf("Done\n");

  break;

  case 'c':

    DemoLogf("DC-DC Converter Voice Band Noise Squelch\n");

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                            108,     // Si3210 Enhancement register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);

    DemoLogf("\nRegister value was %#X\n", regValueL);
    // Bit 1 used to enable audio band noise squelch
    regValueL = regValueL & 0x02;

    if(regValueL) {
      DemoLogf("Voice band squelch enabled\n");
    }
    else {
      DemoLogf("Voice band squelch disabled\n");
    }

    DemoLogf("Enter (y) to enable (n) to disable squelch \n> ");
    OsaConsoleGetStr(strBuf);

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                            108,     // Si3210 Enhancement register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);


    if((strBuf[0] == 'y') || (strBuf[0] == 'Y')) {
      regValueL = regValueL | 0x02;
    }
    else
    if((strBuf[0] == 'n') || (strBuf[0] == 'N')) {
      regValueL = regValueL & ~(0x02);
    }
    else {
      DemoLogf("Software application error. Unsupported option.\n");
      return 0;
    }

    result = OtxDrvWriteReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                             108,     // Si3210 Enhancement register
                             regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);
    
    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                            108,     // Si3210 Enhancement register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);

    DemoLogf("Register value is %#X\n", regValueL);

    DemoLogf("Done\n");

  break;


  case 'd':

    DemoLogf("DC-DC Converter Voice Band Noise Squelch Responce Time\n");

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                            108,     // Si3210 Enhancement register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);

    DemoLogf("\nRegister value was %#X\n", regValueL);
    // Bit 1 used to enable audio band noise squelch
    regValueL = regValueL & 0x20;

    if(regValueL) {
      DemoLogf("Voice band squelch ENHANCED Algorithm\n");
    }
    else {
      DemoLogf("Voice band squelch NORMAL Algorithm\n");
    }

    DemoLogf("Enter (y) to enhanced (n) to Normal squelch \n> ");
    OsaConsoleGetStr(strBuf);

    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                            108,     // Si3210 Enhancement register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);


    if((strBuf[0] == 'y') || (strBuf[0] == 'Y')) {
      regValueL = regValueL | 0x20;
    }
    else
    if((strBuf[0] == 'n') || (strBuf[0] == 'N')) {
      regValueL = regValueL & ~(0x20);
    }
    else {
      DemoLogf("Software application error. Unsupported option.\n");
      return 0;
    }

    result = OtxDrvWriteReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                             108,     // Si3210 Enhancement register
                             regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);
    
    result = OtxDrvReadReg8(g_hPhysDevs.hSlicLi[g_liConNumber_1],
                            108,     // Si3210 Enhancement register
                            &regValueL);
    DemoCheckRc(g_hPhysDevs.hSlicLi[g_liConNumber_1], result);

    DemoLogf("Register value is %#X\n", regValueL);

    DemoLogf("Done\n");

  break;


  case 'Q':
    return 1;
  break;

  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
    // Select current Li
    {
      OTX_INT32 nLi = commandChar - '0';
      
      g_liConNumber_1 = nLi;
      DemoLogf("SLIC #%d Selected ....\n", g_liConNumber_1);
    }
  break;

  default:
    DemoLogf("\nCmd Si3210 Audio Receive Path Configuration\n");
    DemoLogf("The following changes can be made on an active line.\n");
    DemoLogf("===  ========\n");
    DemoLogf(" Q   Return to Main Menu\n");
    DemoLogf(" 0-5 Select active SLIC. #%d is now active\n", g_liConNumber_1);
    DemoLogf(" A   Set Transmit Audio Gain\n");
    DemoLogf(" a   Set Receive Audio Gain\n");
    DemoLogf(" b   Set Receive Hybrid Gain\n");
    DemoLogf(" c   Set Receive DC-DC Converter Noise Squelch\n");
    DemoLogf(" d   Set Receive DC-DC Converter Noise Squelch Response\n");
    DemoLogf("\n > ");
  break;
  }
  
  return 0;
}




